import logging
import pathlib

import pytest

from competitive_verifier.models.problem import TestCaseFile as TstCaseFile
from competitive_verifier.oj.problem import LocalProblem


@pytest.mark.allow_mkdir
def test_local_problem(caplog: pytest.LogCaptureFixture, testtemp: pathlib.Path):
    NOT_FOUND_TUPLE = (
        "competitive_verifier.oj.file",
        logging.WARNING,
        "no cases found",
    )
    DANGLING_OUTPUT_TUPLE = (
        "competitive_verifier.oj.file",
        logging.WARNING,
        "dangling output case",
    )

    problem = LocalProblem(testtemp)
    assert not problem.download_system_cases()
    assert caplog.record_tuples == [NOT_FOUND_TUPLE]
    caplog.clear()
    assert list(problem.iter_system_cases()) == []
    assert caplog.record_tuples == [NOT_FOUND_TUPLE]
    caplog.clear()

    (testtemp / "top.in").touch()
    (testtemp / "top.out").touch()
    (testtemp / "other.tmp").touch()

    (testtemp / "subdir").mkdir()
    (testtemp / "subdir" / "ss.in").touch()
    (testtemp / "subdir" / "ss.out").touch()
    (testtemp / "subdir" / "other.tmp").touch()

    assert problem.download_system_cases()
    assert list(problem.iter_system_cases()) == [
        TstCaseFile(
            name="subdir/ss",
            input_path=testtemp / "subdir/ss.in",
            output_path=testtemp / "subdir/ss.out",
        ),
        TstCaseFile(
            name="top",
            input_path=testtemp / "top.in",
            output_path=testtemp / "top.out",
        ),
    ]
    assert caplog.record_tuples == []

    (testtemp / "subdir" / "only_in.in").touch()
    (testtemp / "subdir" / "only_out.out").touch()
    assert problem.download_system_cases()
    assert list(problem.iter_system_cases()) == [
        TstCaseFile(
            name="subdir/ss",
            input_path=testtemp / "subdir/ss.in",
            output_path=testtemp / "subdir/ss.out",
        ),
        TstCaseFile(
            name="top",
            input_path=testtemp / "top.in",
            output_path=testtemp / "top.out",
        ),
    ]
    assert caplog.record_tuples == [DANGLING_OUTPUT_TUPLE, DANGLING_OUTPUT_TUPLE]



import sys
from contextlib import contextmanager
from logging import (
    NOTSET,
    WARNING,
    Handler,
    basicConfig,
)
from typing import TextIO

import colorlog
from colorama import Fore, Style

from competitive_verifier import github


def configure_stderr_logging(
    default_level: int | None = None,
) -> None:  # pragma: no cover
    colorlog_handler = colorlog.StreamHandler(sys.stderr)
    colorlog_handler.setLevel(default_level or WARNING)
    colorlog_handler.setFormatter(
        colorlog.ColoredFormatter(
            "%(log_color)s%(levelname)s%(reset)s:%(name)s:%(lineno)d:%(message)s"
        )
    )
    handlers: list[Handler] = [colorlog_handler]

    basicConfig(
        level=NOTSET,
        handlers=handlers,
    )


def _console_group(category: str, *, title: str, file: TextIO | None):
    print(
        (
            f"<------------- {Fore.CYAN}{category}:"
            f"{Fore.YELLOW}{title}"
            f"{Style.RESET_ALL} ------------->"
        ),
        file=file,
    )


@contextmanager
def group(title: str, *, stream: TextIO | None = None):
    if stream is None:
        stream = sys.stderr
    if github.env.is_in_github_actions():
        try:
            github.begin_group(title, stream=stream)
            yield
        finally:
            github.end_group(stream=stream)
    else:
        try:
            _console_group(" Start group", title=title, file=stream)
            yield
        finally:
            _console_group("Finish group", title=title, file=stream)



from competitive_verifier.oj.verify.languages.cplusplus import (
    CPlusPlusLanguage,
    OjVerifyCPlusPlusConfig,
)
from competitive_verifier.oj.verify.languages.go import GoLanguage, OjVerifyGoConfig
from competitive_verifier.oj.verify.languages.haskell import (
    HaskellLanguage,
    OjVerifyHaskellConfig,
)
from competitive_verifier.oj.verify.languages.java import (
    JavaLanguage,
    OjVerifyJavaConfig,
)
from competitive_verifier.oj.verify.languages.nim import NimLanguage, OjVerifyNimConfig
from competitive_verifier.oj.verify.languages.python import PythonLanguage
from competitive_verifier.oj.verify.languages.ruby import (
    OjVerifyRubyConfig,
    RubyLanguage,
)
from competitive_verifier.oj.verify.languages.rust import (
    OjVerifyRustConfig,
    RustLanguage,
)
from competitive_verifier.oj.verify.languages.user_defined import UserDefinedLanguage

__all__ = [
    "CPlusPlusLanguage",
    "GoLanguage",
    "HaskellLanguage",
    "JavaLanguage",
    "NimLanguage",
    "OjVerifyCPlusPlusConfig",
    "OjVerifyGoConfig",
    "OjVerifyHaskellConfig",
    "OjVerifyJavaConfig",
    "OjVerifyNimConfig",
    "OjVerifyRubyConfig",
    "OjVerifyRustConfig",
    "PythonLanguage",
    "RubyLanguage",
    "RustLanguage",
    "UserDefinedLanguage",
]



from typing import Literal, TextIO

from . import env


def print_error(
    message: str,
    *,
    title: str | None = None,
    file: str | None = None,
    col: int | None = None,
    end_column: int | None = None,
    line: int | None = None,
    end_line: int | None = None,
    stream: TextIO | None = None,
) -> None:
    _print_github(
        "error",
        message,
        title=title,
        file=file,
        col=col,
        end_column=end_column,
        line=line,
        end_line=end_line,
        stream=stream,
    )


def _print_github(
    command: Literal["error", "warning", "debug"],
    message: str,
    *,
    title: str | None = None,
    file: str | None = None,
    col: int | None = None,
    end_column: int | None = None,
    line: int | None = None,
    end_line: int | None = None,
    stream: TextIO | None = None,
) -> None:
    """Print Github Actions style message.

    https://docs.github.com/en/actions/using-workflows/workflow-commands-for-github-actions

    Args:
        command: Command name (e.g., "error", "warning", "debug")
        message: Message body
    Keyword Args:
        title: Custom title
        file: Filename
        col: Column number, starting at 1
        end_column: End column number
        line: Line number, starting at 1
        end_line: End line number
        stream: Output stream
    """
    if env.is_in_github_actions():
        annotation = ",".join(
            f"{name}={value}"
            for name, value in (
                ("title", title),
                ("file", file),
                ("col", col),
                ("endColumn", end_column),
                ("line", line),
                ("endLine", end_line),
            )
            if value is not None
        )

        print(f"::{command} {annotation}::{message}", file=stream)


def begin_group(title: str, *, stream: TextIO | None = None):
    print(f"::group::{title}", file=stream)


def end_group(*, stream: TextIO | None = None):
    print("::endgroup::", file=stream)



import os
import pathlib
import tempfile
from collections.abc import Generator

import pytest
from pytest_mock import MockerFixture


@pytest.fixture
def mock_perf_counter(mocker: MockerFixture, request: pytest.FixtureRequest):
    if not hasattr(request, "param") or not request.param:
        values = [float(i) for i in range(1000)]
    else:
        assert isinstance(request.param, list)
        values = request.param  # pyright: ignore[reportUnknownMemberType, reportUnknownVariableType]

    mocker.patch("time.perf_counter", side_effect=values)


@pytest.fixture
def mockenv(mocker: MockerFixture, request: pytest.FixtureRequest):
    mocker.patch.dict(os.environ, request.param or {}, clear=True)


@pytest.fixture(autouse=True)
def mock_mkdir(mocker: MockerFixture, request: pytest.FixtureRequest):
    if "allow_mkdir" in request.keywords:
        return None
    return mocker.patch(
        "pathlib.Path.mkdir", side_effect=RuntimeError("mkdir is not allowed")
    )


@pytest.fixture
def testtemp(monkeypatch: pytest.MonkeyPatch) -> Generator[pathlib.Path, None, None]:
    with tempfile.TemporaryDirectory() as d:
        monkeypatch.chdir(d)
        yield pathlib.Path(d).resolve()
        monkeypatch.undo()



import os
import sys
from io import StringIO

import pytest
from pytest_mock import MockerFixture

from competitive_verifier import github
from competitive_verifier.log import group


class TestGitHubPrint:
    def test_print_error(
        self,
        capsys: pytest.CaptureFixture[str],
        mocker: MockerFixture,
    ):
        mocker.patch.dict(os.environ, {"GITHUB_ACTIONS": ""}, clear=True)
        github.print_error("simple_message")
        out, err = capsys.readouterr()
        assert out == ""
        assert err == ""

        mocker.patch.dict(os.environ, {"GITHUB_ACTIONS": "true"}, clear=True)
        github.print_error("simple_message")
        out, err = capsys.readouterr()
        assert out == "::error ::simple_message\n"
        assert err == ""

    def test_print_error_github_actions(
        self,
        capsys: pytest.CaptureFixture[str],
        mocker: MockerFixture,
    ):
        mocker.patch.dict(os.environ, {"GITHUB_ACTIONS": "true"}, clear=True)
        github.print_error(
            "many_message",
            title="TestTitle",
            file="foo.txt",
            col=10,
            end_column=20,
            line=100,
            end_line=101,
            stream=sys.stderr,
        )
        out, err = capsys.readouterr()
        assert out == ""
        assert (
            err
            == "::error title=TestTitle,file=foo.txt,col=10,endColumn=20,line=100,endLine=101::many_message\n"
        )


class TestLogGroup:
    def test_log_group_github_actions(
        self, mocker: MockerFixture, capsys: pytest.CaptureFixture[str]
    ):
        mocker.patch.dict(os.environ, {"GITHUB_ACTIONS": "true"}, clear=True)
        with group("TestTitle"):
            out, err = capsys.readouterr()
            assert out == ""
            assert err == "::group::TestTitle\n"

        out, err = capsys.readouterr()
        assert out == ""
        assert err == "::endgroup::\n"

    def test_log_group_github_actions_stream(self, mocker: MockerFixture):
        mocker.patch.dict(os.environ, {"GITHUB_ACTIONS": "true"}, clear=True)
        with StringIO() as fp:
            with group("TestTitle", stream=fp):
                fp.seek(0)
                assert fp.read() == "::group::TestTitle\n"
                fp.seek(0)
                fp.truncate(0)
            fp.seek(0)
            assert fp.read() == "::endgroup::\n"

    def test_log_group_console(
        self, capsys: pytest.CaptureFixture[str], mocker: MockerFixture
    ):
        mocker.patch.dict(os.environ, {"GITHUB_ACTIONS": ""}, clear=True)
        with group("TestTitle"):
            out, err = capsys.readouterr()
            assert out == ""
            assert (
                err
                == "<------------- \x1b[36m Start group:\x1b[33mTestTitle\x1b[0m ------------->\n"
            )

        out, err = capsys.readouterr()
        assert out == ""
        assert (
            err
            == "<------------- \x1b[36mFinish group:\x1b[33mTestTitle\x1b[0m ------------->\n"
        )

    def test_log_group_console_stream(self, mocker: MockerFixture):
        mocker.patch.dict(os.environ, {"GITHUB_ACTIONS": ""}, clear=True)
        with StringIO() as fp:
            with group("TestTitle", stream=fp):
                fp.seek(0)
                assert (
                    fp.read()
                    == "<------------- \x1b[36m Start group:\x1b[33mTestTitle\x1b[0m ------------->\n"
                )
                fp.seek(0)
                fp.truncate(0)
            fp.seek(0)
            assert (
                fp.read()
                == "<------------- \x1b[36mFinish group:\x1b[33mTestTitle\x1b[0m ------------->\n"
            )



import contextlib
import math
import os
import pathlib
import platform
import shlex
import signal
import subprocess
import sys
import tempfile
import time
from dataclasses import dataclass
from logging import getLogger
from typing import BinaryIO

from competitive_verifier.models import (
    JudgeStatus,
    ResultStatus,
    TestCaseProvider,
    TestcaseResult,
    VerifcationTimeoutError,
    VerificationResult,
)

from . import gnu
from .format import Printer, StatusCounter, green, red

logger = getLogger(__name__)


class _ExecError(Exception):
    pass


@dataclass
class OjExecInfo:
    answer: bytes | None
    """The standard output of the executed command"""
    elapsed: float
    """The elapsed time of the executed command in seconds"""
    memory: float | None
    """The maximum memory usage of the executed command in megabytes"""
    returncode: int | None
    """The returncode of the executed command"""


def measure_command(
    command: list[str] | str,
    *,
    env: dict[str, str] | None = None,
    stdin: BinaryIO | int | None = None,
    input: bytes | None = None,  # noqa: A002
    timeout: float | None = None,
    gnu_time: bool = False,
) -> OjExecInfo:
    if input is not None:
        if stdin is not None:
            raise ValueError(
                stdin, "stdin and input cannot be specified at the same time"
            )
        stdin = subprocess.PIPE

    if isinstance(command, str):
        command = shlex.split(command)
    with gnu.GnuTimeWrapper(enabled=gnu_time) as gw:
        command = gw.get_command(command)
        begin = time.perf_counter()

        # We need kill processes called from the "time" command using process groups. Without this, orphans spawn. see https://github.com/kmyk/online-judge-tools/issues/640
        start_new_session = gnu.time_command() is not None and os.name == "posix"

        try:
            if env:
                env = os.environ | env
            proc = subprocess.Popen(
                command,
                stdin=stdin,
                stdout=subprocess.PIPE,
                env=env,
                stderr=sys.stderr,
                start_new_session=start_new_session,
            )
        except FileNotFoundError as e:
            logger.exception("exec:No such file or directory: %s", command)
            raise _ExecError from e
        except PermissionError as e:
            logger.exception("exec:Permission denied: %s", command)
            raise _ExecError from e
        answer: bytes | None = None
        try:
            answer, _ = proc.communicate(input=input, timeout=timeout)
        except subprocess.TimeoutExpired:
            pass
        finally:
            if start_new_session:
                with contextlib.suppress(ProcessLookupError):
                    os.killpg(os.getpgid(proc.pid), signal.SIGTERM)
            else:
                proc.terminate()

        end = time.perf_counter()
        memory: float | None = None
        report = gw.get_report()
        if report:
            logger.debug("GNU time says:\n%s", report)
            if report.splitlines()[-1].isdigit():
                memory = int(report.splitlines()[-1]) / 1000
        return OjExecInfo(
            answer=answer,
            elapsed=end - begin,
            memory=memory,
            returncode=proc.returncode,
        )


@dataclass
class OjTestArguments:
    """Parameters for oj-test command.

    Port of onlinejudge_command.subcommand.test.add_subparser.
    """

    command: str | list[str]
    problem: TestCaseProvider
    tle: float | None
    mle: float | None
    error: float | None
    env: dict[str, str] | None = None
    deadline: float = float("inf")


@dataclass
class OjTestcaseResult:
    name: str
    """A name of the test case."""
    input: pathlib.Path
    """A input of the test case."""
    answer: bytes
    """A output of the test case."""

    status: JudgeStatus
    elapsed: float
    exitcode: int | None

    memory: float | None = None
    expected: pathlib.Path | None = None
    """A expected output of the test case."""

    def __post_init__(self):
        if not isinstance(self.exitcode, int):
            self.exitcode = None

    def __str__(self) -> str:
        p = [
            f"{self.name}: {green('AC')}"
            if self.status == JudgeStatus.AC
            else f"{self.name}: {red(self.status.name)}",
            f"time: {self.elapsed:f} sec",
            f"memory: {self.memory:f} MB" if self.memory is not None else None,
            f"return code: {self.exitcode}" if self.exitcode else None,
        ]

        return ", ".join(filter(None, p))

    def log(self):
        match self.status:
            case JudgeStatus.AC:
                if self.expected is None and self.answer:
                    self._log_answer()
            case JudgeStatus.RE | JudgeStatus.TLE:
                self._log_input()
                self._log_expected()
            case _:
                self._log_input()
                self._log_answer()
                self._log_expected()
        logger.info(self)

    def _log_input(self) -> None:
        with self.input.open("rb") as fp:
            logger.info("%s:input:\n%s", self.name, Printer(fp))

    def _log_expected(self) -> None:
        if self.expected:
            with self.expected.open("rb") as fp:
                logger.info("%s:expected:\n%s", self.name, Printer(fp))
        else:
            logger.info("%s:expected:\n%s", self.name, Printer(""))

    def _log_answer(self) -> None:
        logger.info("%s:answer:\n%s", self.name, Printer(self.answer))


@dataclass
class OjTestResult:
    is_success: bool

    elapsed: float

    slowest: float
    """max time [seconds]
    """

    heaviest: float
    """max memory [MB]
    """

    testcases: list[OjTestcaseResult]


class SpecialJudge:
    def __init__(
        self,
        judge_command: str,
        *,
        input_path: pathlib.Path,
        expected_output_path: pathlib.Path | None,
    ):
        self.judge_command = judge_command  # already quoted and joined command
        self.input_path = input_path
        self.expected_output_path = expected_output_path

    def run(self, actual_output: bytes, _: bytes) -> bool:
        with tempfile.TemporaryDirectory() as tempdir:
            actual_output_path = pathlib.Path(tempdir) / "actual.out"
            with actual_output_path.open("wb") as fh:
                fh.write(actual_output)

            # if you use shlex.quote, it fails on Windows. why?
            command = " ".join(
                [
                    self.judge_command,  # already quoted and joined command
                    str(self.input_path.resolve()),
                    str(actual_output_path.resolve()),
                    str(
                        self.expected_output_path.resolve()
                        if self.expected_output_path is not None
                        else ""
                    ),
                ]
            )

            logger.debug("$ %s", command)
            info = measure_command(command)
        logger.debug("judge's output:\n%s", Printer(info.answer or ""))
        return info.returncode == 0


def _try_parse_float(value: bytes) -> float | None:
    try:
        return float(value)
    except ValueError:
        return None


def _equal_or_closed_float(actual: bytes, expected: bytes, *, error: float) -> bool:
    if actual == expected:
        return True

    x = _try_parse_float(actual)
    y = _try_parse_float(expected)

    return (
        x is not None
        and y is not None
        and math.isclose(x, y, rel_tol=error, abs_tol=error)
    )


def compare_answer(actual: bytes, expected: bytes, *, error: float | None) -> bool:
    """Compare two byte strings.

    Args:
        actual (bytes): Actual output
        expected (bytes): Expected output
        error (float | None): Margin of error
    Returns:
        bool: True if they are considered equal
    """
    if error is not None and error > 1:
        logger.warning("the tolerance is too large: relative = %s", error)

    actual = actual.replace(b"\r\n", b"\n")
    expected = expected.replace(b"\r\n", b"\n")

    # match
    if actual == expected:
        return True

    try:
        if error is None:
            actual_words = actual.split()
            expected_words = expected.split()
            if all(x == y for x, y in zip(actual_words, expected_words, strict=True)):
                logger.warning("This was AC if spaces and newlines were ignored.")
            return False

        actual_lines = actual.rstrip(b"\n").split(b"\n")
        expected_lines = expected.rstrip(b"\n").split(b"\n")

        for actual_line, expected_line in zip(
            actual_lines, expected_lines, strict=True
        ):
            actual_words = actual_line.split()
            expected_words = expected_line.split()

            for x, y in zip(actual_words, expected_words, strict=True):
                if not _equal_or_closed_float(x, y, error=error):
                    return False
    except ValueError:
        return False

    return True


def check_output(
    *,
    answer: bytes,
    error: float | None,
    test_input_path: pathlib.Path,
    test_output_path: pathlib.Path | None,
    judge_command: str | None,
) -> bool | None:
    if test_output_path is None and not judge_command:
        return None
    if test_output_path is not None:
        with test_output_path.open("rb") as outf:
            expected = outf.read()
    else:
        # only if --judge option
        expected = b""
        logger.warning("expected output is not found")
    if judge_command is not None:
        return SpecialJudge(
            judge_command,
            input_path=test_input_path,
            expected_output_path=test_output_path,
        ).run(answer, expected)

    return compare_answer(answer, expected, error=error)


def determine_status(
    *,
    exitcode: int | None,
    memory: float | None,
    mle: float | None,
    match_result: bool | None,
) -> JudgeStatus:
    if exitcode is None:
        return JudgeStatus.TLE
    if memory is not None and mle is not None and memory > mle:
        return JudgeStatus.MLE
    if exitcode != 0:
        return JudgeStatus.RE
    if match_result is not None and not match_result:
        return JudgeStatus.WA
    return JudgeStatus.AC


def single_case(
    test_name: str,
    test_input_path: pathlib.Path,
    test_output_path: pathlib.Path | None,
    *,
    args: OjTestArguments,
) -> OjTestcaseResult:
    try:
        logger.info("%s: start", test_name)

        # run the binary
        with test_input_path.open("rb") as infp:
            info = measure_command(
                args.command,
                env=args.env,
                stdin=infp,
                timeout=args.tle,
                gnu_time=True,
            )
            answer = info.answer or b""
            elapsed: float = info.elapsed
            memory: float | None = info.memory

        match_result = check_output(
            answer=answer,
            error=args.error,
            test_input_path=test_input_path,
            test_output_path=test_output_path,
            judge_command=str(args.problem.checker) if args.problem.checker else None,
        )
        status = determine_status(
            exitcode=info.returncode,
            memory=memory,
            mle=args.mle,
            match_result=match_result,
        )

        result = OjTestcaseResult(
            name=test_name,
            input=test_input_path,
            expected=test_output_path,
            answer=answer,
            status=status,
            exitcode=info.returncode,
            elapsed=elapsed,
            memory=memory,
        )
    except _ExecError:
        return OjTestcaseResult(
            name=test_name,
            input=test_input_path,
            expected=test_output_path,
            answer=b"",
            status=JudgeStatus.RE,
            exitcode=255,
            elapsed=0,
            memory=None,
        )
    else:
        result.log()
        return result


def _run(args: OjTestArguments) -> OjTestResult:
    # check wheather GNU time is available
    if gnu.time_command() is None:
        if platform.system() == "Darwin":
            logger.info(
                "%s: You can install GNU time with: $ brew install gnu-time",
            )
        if args.mle is not None:
            raise RuntimeError("--mle is used but GNU time does not exist")

    tests = list(args.problem.iter_system_cases())

    # run tests
    history: list[OjTestcaseResult] = []
    for t in tests:
        if time.perf_counter() > args.deadline:
            raise VerifcationTimeoutError

        history.append(single_case(t.name, t.input_path, t.output_path, args=args))

    # summarize
    elapsed: float = 0.0
    slowest: float = -1.0
    slowest_name = ""
    heaviest: float = -1.0
    heaviest_name = ""
    counter = StatusCounter()
    for result in history:
        counter[result.status] += 1
        elapsed += result.elapsed
        if slowest < result.elapsed:
            slowest = result.elapsed
            slowest_name = result.name
        if result.memory is not None and heaviest < result.memory:
            heaviest = result.memory
            heaviest_name = result.name

    # print the summary
    logger.info("slowest: %f sec  (for %s)", slowest, slowest_name)
    if heaviest >= 0:
        logger.info("max memory: %f MB  (for %s)", heaviest, heaviest_name)

    is_success = counter[JudgeStatus.AC] == len(tests)
    if is_success:
        logger.info("%s %d cases", green("SUCCESS"), len(tests))
    else:
        logger.info("%s %s / %d cases", red("FAILURE"), counter, len(tests))

    # return the result
    return OjTestResult(
        is_success=is_success,
        slowest=slowest,
        elapsed=elapsed,
        heaviest=heaviest,
        testcases=history,
    )


def main(
    *,
    problem: TestCaseProvider,
    command: str | list[str],
    env: dict[str, str] | None,
    tle: float | None,
    mle: float | None,
    error: float | None,
    deadline: float = float("inf"),
) -> VerificationResult:
    args = OjTestArguments(
        command=command,
        problem=problem,
        env=env,
        tle=tle,
        mle=mle,
        error=error,
        deadline=deadline,
    )
    result = _run(args)
    return VerificationResult(
        status=ResultStatus.SUCCESS if result.is_success else ResultStatus.FAILURE,
        elapsed=result.elapsed,
        slowest=result.slowest,
        heaviest=result.heaviest,
        testcases=[
            TestcaseResult(
                name=case.name,
                elapsed=case.elapsed,
                memory=case.memory,
                status=case.status,
            )
            for case in result.testcases
        ],
    )



import pathlib
from typing import NamedTuple

import pytest

from competitive_verifier.util import resolve_relative_or_abs_path


class DirStructure(NamedTuple):
    name: str
    items: "list[str|DirStructure]"


@pytest.mark.allow_mkdir
def test_resolve_relative_or_abs_path(
    monkeypatch: pytest.MonkeyPatch,
    testtemp: pathlib.Path,
):
    def setup(cwd: pathlib.Path, d: DirStructure):
        dd = cwd / d.name
        dd.mkdir()
        for c in d.items:
            if isinstance(c, str):
                (dd / c).touch()
            else:
                setup(dd, c)

    setup(
        testtemp,
        DirStructure(
            "root",
            [
                "outer",
                DirStructure(
                    "dir1",
                    ["file1.txt", DirStructure("subdirA", ["a1.cb1.c"])],
                ),
                DirStructure(
                    "dir2",
                    ["file1.txt", DirStructure("subdirB", ["a1.cb1.c"])],
                ),
            ],
        ),
    )
    monkeypatch.chdir(testtemp / "root")

    assert (
        resolve_relative_or_abs_path(str(testtemp / "outer"), basedir=testtemp / "root")
        is None
    )

    assert resolve_relative_or_abs_path(
        "./file1.txt", basedir=testtemp / "root/dir1"
    ) == pathlib.Path("dir1/file1.txt")

    assert resolve_relative_or_abs_path(
        "../dir2/file1.txt", basedir=testtemp / "root/dir1"
    ) == pathlib.Path("dir2/file1.txt")

    assert (
        resolve_relative_or_abs_path("./file2.txt", basedir=testtemp / "root/dir1")
        is None
    )

    assert (
        resolve_relative_or_abs_path(
            "../dir2/file2.txt", basedir=testtemp / "root/dir1"
        )
        is None
    )

    assert resolve_relative_or_abs_path(
        "//dir1/file1.txt", basedir=testtemp / "root/dir1"
    ) == pathlib.Path("dir1/file1.txt")

    assert resolve_relative_or_abs_path(
        "//dir2/file1.txt", basedir=testtemp / "root/dir1"
    ) == pathlib.Path("dir2/file1.txt")

    assert (
        resolve_relative_or_abs_path("//file2.txt", basedir=testtemp / "root/dir1")
        is None
    )

    assert resolve_relative_or_abs_path(
        "file1.txt", basedir=testtemp / "root/dir1"
    ) == pathlib.Path("dir1/file1.txt")

    assert resolve_relative_or_abs_path(
        str(testtemp / "root/dir1/file1.txt"), basedir=testtemp / "root/dir1"
    ) == pathlib.Path("dir1/file1.txt")



from pydantic import Field

from competitive_verifier.models import ShellCommand, ShellCommandLike
from competitive_verifier.oj.verify.models import OjVerifyUserDefinedConfig

from .user_defined import UserDefinedLanguage


class OjVerifyGoConfig(OjVerifyUserDefinedConfig):
    execute: ShellCommandLike = Field(
        default_factory=lambda: ShellCommand(
            command=["go", "run", "{basedir}/{path}"],
            env={"GO111MODULE": "off"},
        ),
    )


class GoLanguage(UserDefinedLanguage):
    def __init__(self, *, config: OjVerifyGoConfig | None):
        super().__init__(extension="go", config=config or OjVerifyGoConfig())



import logging
import pathlib
from typing import Any, NamedTuple

import pytest
from pytest_mock import MockerFixture

from competitive_verifier import oj
from competitive_verifier.models import JudgeStatus, Problem
from competitive_verifier.oj.oj_test import (
    OjExecInfo,
    OjTestArguments,
    OjTestcaseResult,
    SpecialJudge,
    compare_answer,
    single_case,
)
from competitive_verifier.oj.problem import (
    AOJProblem,
    LibraryCheckerProblem,
    YukicoderProblem,
)

OJ_TEST_MODULE = "competitive_verifier.oj.oj_test"


@pytest.fixture
def mock_measure(
    mocker: MockerFixture,
    request: pytest.FixtureRequest,
):
    assert isinstance(request.param, OjExecInfo)

    mocker.patch(
        "competitive_verifier.oj.oj_test.measure_command",
        return_value=request.param,
    )
    return request.param


@pytest.fixture
def mock_generate_test_cases(mocker: MockerFixture):
    mocker.patch.object(
        LibraryCheckerProblem,
        "generate_test_cases",
    )


@pytest.fixture
def mock_judge(
    mocker: MockerFixture,
    request: pytest.FixtureRequest,
) -> Problem:
    assert request.param is None or isinstance(request.param, bool)

    if request.param is not None:
        mocker.patch.object(SpecialJudge, "run", return_value=request.param)
        mocker.patch.object(
            LibraryCheckerProblem,
            "checker",
            return_value=pathlib.Path("/anywhere/mockcheck"),
            new_callable=mocker.PropertyMock,
        )
        return LibraryCheckerProblem(problem_id="aplusb")
    return AOJProblem(problem_id="1")


class SingleCaseParams(NamedTuple):
    name: str
    inbytes: bytes
    mock_measure: OjExecInfo
    expected: OjTestcaseResult
    expected_log: list[tuple[str, int, str]]
    outbytes: bytes | None = None
    error: float | None = None
    mle: float | None = None
    mock_judge: bool | None = None


def make_result(
    name: str,
    elapsed: float,
    exitcode: int | None,
    memory: float | None,
    status: JudgeStatus,
) -> OjTestcaseResult:
    return OjTestcaseResult(
        name=name,
        elapsed=elapsed,
        exitcode=exitcode,
        memory=memory,
        status=status,
        input=pathlib.Path(),
        answer=b"",
    )


def log_output(msg: str, *, module: str = OJ_TEST_MODULE, level: int = logging.INFO):
    return (module, level, msg)


test_single_case_params: list[SingleCaseParams] = [
    SingleCaseParams(
        name="default",
        inbytes=b"abc\n",
        outbytes=b"ABC\n",
        mock_measure=OjExecInfo(
            answer=b"ABC\n", elapsed=1.25, memory=None, returncode=0
        ),
        expected=make_result(
            name="default",
            elapsed=1.25,
            exitcode=0,
            memory=None,
            status=JudgeStatus.AC,
        ),
        expected_log=[
            log_output("default: start"),
            log_output("default: \x1b[32mAC\x1b[39m, time: 1.250000 sec"),
        ],
    ),
    SingleCaseParams(
        name="no_output",
        inbytes=b"abc\n",
        mock_measure=OjExecInfo(
            answer=b"ABC\n", elapsed=1.25, memory=None, returncode=0
        ),
        expected=make_result(
            name="no_output",
            elapsed=1.25,
            exitcode=0,
            memory=None,
            status=JudgeStatus.AC,
        ),
        expected_log=[
            log_output("no_output: start"),
            log_output("no_output:answer:\n\x1b[1mABC\x1b[0m\x1b[2m\\n\x1b[0m"),
            log_output("no_output: \x1b[32mAC\x1b[39m, time: 1.250000 sec"),
        ],
    ),
    SingleCaseParams(
        name="Nones",
        inbytes=b"",
        mock_measure=OjExecInfo(answer=None, elapsed=1.25, memory=None, returncode=0),
        expected=make_result(
            name="Nones",
            elapsed=1.25,
            exitcode=0,
            memory=None,
            status=JudgeStatus.AC,
        ),
        expected_log=[
            log_output("Nones: start"),
            log_output("Nones: \x1b[32mAC\x1b[39m, time: 1.250000 sec"),
        ],
    ),
    SingleCaseParams(
        name="memory",
        inbytes=b"",
        mock_measure=OjExecInfo(
            answer=b"19\n20", elapsed=1.25, memory=10.7, returncode=0
        ),
        expected=make_result(
            name="memory",
            elapsed=1.25,
            exitcode=0,
            memory=10.7,
            status=JudgeStatus.AC,
        ),
        expected_log=[
            log_output("memory: start"),
            log_output(
                msg="memory:answer:\n"
                "\x1b[1m19\x1b[0m\x1b[2m\\n\x1b[0m\x1b[1m20\x1b[0m\x1b[2m(no trailing newline)\x1b[0m",
            ),
            log_output(
                "memory: \x1b[32mAC\x1b[39m, time: 1.250000 sec, memory: 10.700000 MB"
            ),
        ],
    ),
    SingleCaseParams(
        name="spaces",
        inbytes=b"foo \t bar\nbaz \r\nhoge\nfuga\n",
        outbytes=b"1\n",
        mock_measure=OjExecInfo(answer=b"2\n", elapsed=1.25, memory=None, returncode=0),
        expected=make_result(
            name="spaces",
            elapsed=1.25,
            exitcode=0,
            memory=None,
            status=JudgeStatus.WA,
        ),
        expected_log=[
            log_output("spaces: start"),
            log_output(
                "spaces:input:\n"
                "\x1b[1mfoo\x1b[0m\x1b[2m_\\t_\x1b[0m\x1b[1mbar\x1b[0m\x1b[2m\\n"
                "\x1b[0m\x1b[1mbaz\x1b[0m\x1b[2m_\\r\x1b[0m\x1b[2m(trailing whitespace)\x1b[0m\x1b[2m\\n"
                "\x1b[0m\x1b[1mhoge\x1b[0m\x1b[2m\\n\x1b[0m\x1b[1mfuga\x1b[0m\x1b[2m\\n\x1b[0m",
            ),
            log_output("spaces:answer:\n\x1b[1m2\x1b[0m\x1b[2m\\n\x1b[0m"),
            log_output("spaces:expected:\n\x1b[1m1\x1b[0m\x1b[2m\\n\x1b[0m"),
            log_output("spaces: \x1b[31mWA\x1b[39m, time: 1.250000 sec"),
        ],
    ),
    SingleCaseParams(
        name="empty-lines",
        inbytes=b"\n\n\n\n",
        outbytes=b"1\n",
        mock_measure=OjExecInfo(answer=b"2\n", elapsed=1.25, memory=None, returncode=0),
        expected=make_result(
            name="empty-lines",
            elapsed=1.25,
            exitcode=0,
            memory=None,
            status=JudgeStatus.WA,
        ),
        expected_log=[
            log_output("empty-lines: start"),
            log_output("empty-lines:input:\n\x1b[2m\\n\\n\\n\\n\x1b[0m"),
            log_output("empty-lines:answer:\n\x1b[1m2\x1b[0m\x1b[2m\\n\x1b[0m"),
            log_output("empty-lines:expected:\n\x1b[1m1\x1b[0m\x1b[2m\\n\x1b[0m"),
            log_output("empty-lines: \x1b[31mWA\x1b[39m, time: 1.250000 sec"),
        ],
    ),
    SingleCaseParams(
        name="judge-AC",
        inbytes=b"abc\n",
        outbytes=b"ABC\n",
        mock_judge=True,
        mock_measure=OjExecInfo(
            answer=b"ABC\n", elapsed=1.25, memory=None, returncode=0
        ),
        expected=make_result(
            name="judge-AC",
            elapsed=1.25,
            exitcode=0,
            memory=None,
            status=JudgeStatus.AC,
        ),
        expected_log=[
            log_output("judge-AC: start"),
            log_output("judge-AC: \x1b[32mAC\x1b[39m, time: 1.250000 sec"),
        ],
    ),
    SingleCaseParams(
        name="judge-WA",
        inbytes=b"abc\n",
        outbytes=b"ABC\n",
        mock_judge=False,
        mock_measure=OjExecInfo(
            answer=b"ABC\n", elapsed=1.25, memory=None, returncode=0
        ),
        expected=make_result(
            name="judge-WA",
            elapsed=1.25,
            exitcode=0,
            memory=None,
            status=JudgeStatus.WA,
        ),
        expected_log=[
            log_output("judge-WA: start"),
            log_output("judge-WA:input:\n\x1b[1mabc\x1b[0m\x1b[2m\\n\x1b[0m"),
            log_output("judge-WA:answer:\n\x1b[1mABC\x1b[0m\x1b[2m\\n\x1b[0m"),
            log_output("judge-WA:expected:\n\x1b[1mABC\x1b[0m\x1b[2m\\n\x1b[0m"),
            log_output("judge-WA: \x1b[31mWA\x1b[39m, time: 1.250000 sec"),
        ],
    ),
    SingleCaseParams(
        name="judge-WA-no-expected-out",
        inbytes=b"abc\n",
        mock_judge=False,
        mock_measure=OjExecInfo(
            answer=b"ABC\n", elapsed=1.25, memory=None, returncode=0
        ),
        expected=make_result(
            name="judge-WA-no-expected-out",
            elapsed=1.25,
            exitcode=0,
            memory=None,
            status=JudgeStatus.WA,
        ),
        expected_log=[
            log_output("judge-WA-no-expected-out: start"),
            log_output("expected output is not found", level=logging.WARNING),
            log_output(
                "judge-WA-no-expected-out:input:\n\x1b[1mabc\x1b[0m\x1b[2m\\n\x1b[0m"
            ),
            log_output(
                "judge-WA-no-expected-out:answer:\n\x1b[1mABC\x1b[0m\x1b[2m\\n\x1b[0m"
            ),
            log_output("judge-WA-no-expected-out:expected:\n\x1b[2m(empty)\x1b[0m"),
            log_output(
                "judge-WA-no-expected-out: \x1b[31mWA\x1b[39m, time: 1.250000 sec"
            ),
        ],
    ),
    SingleCaseParams(
        name="check_output-WA",
        inbytes=b"abc\n",
        outbytes=b"ABC",
        mock_measure=OjExecInfo(
            answer=b"ABC\n", elapsed=1.25, memory=None, returncode=0
        ),
        expected=make_result(
            name="check_output-WA",
            elapsed=1.25,
            exitcode=0,
            memory=None,
            status=JudgeStatus.WA,
        ),
        expected_log=[
            log_output("check_output-WA: start"),
            log_output(
                "This was AC if spaces and newlines were ignored.",
                level=logging.WARNING,
            ),
            log_output(
                "check_output-WA:input:\n\x1b[1mabc\x1b[0m\x1b[2m\\n\x1b[0m",
            ),
            log_output("check_output-WA:answer:\n\x1b[1mABC\x1b[0m\x1b[2m\\n\x1b[0m"),
            log_output(
                "check_output-WA:expected:\n\x1b[1mABC\x1b[0m\x1b[2m(no trailing newline)\x1b[0m",
            ),
            log_output("check_output-WA: \x1b[31mWA\x1b[39m, time: 1.250000 sec"),
        ],
    ),
    SingleCaseParams(
        name="check_output-AC",
        inbytes=b"abc\n",
        outbytes=b"ABC\n",
        mock_measure=OjExecInfo(
            answer=b"ABC\n", elapsed=1.25, memory=None, returncode=0
        ),
        expected=make_result(
            name="check_output-AC",
            elapsed=1.25,
            exitcode=0,
            memory=None,
            status=JudgeStatus.AC,
        ),
        expected_log=[
            log_output("check_output-AC: start"),
            log_output("check_output-AC: \x1b[32mAC\x1b[39m, time: 1.250000 sec"),
        ],
    ),
    SingleCaseParams(
        name="error-AC-equals",
        inbytes=b"Pi\n",
        outbytes=b"3.14159265358\n",
        error=1e-11,
        mock_measure=OjExecInfo(
            answer=b"3.14159265358\n", elapsed=1.25, memory=None, returncode=0
        ),
        expected=make_result(
            name="error-AC-equals",
            elapsed=1.25,
            exitcode=0,
            memory=None,
            status=JudgeStatus.AC,
        ),
        expected_log=[
            log_output("error-AC-equals: start"),
            log_output("error-AC-equals: \x1b[32mAC\x1b[39m, time: 1.250000 sec"),
        ],
    ),
    SingleCaseParams(
        name="error-AC-long",
        inbytes=b"Pi\n",
        outbytes=b"3.141592653589793238462643383279502884197169399375105820974944592307816406286\n",
        error=1e-11,
        mock_measure=OjExecInfo(
            answer=b"3.14159265358\n", elapsed=1.25, memory=None, returncode=0
        ),
        expected=make_result(
            name="error-AC-long",
            elapsed=1.25,
            exitcode=0,
            memory=None,
            status=JudgeStatus.AC,
        ),
        expected_log=[
            log_output("error-AC-long: start"),
            log_output("error-AC-long: \x1b[32mAC\x1b[39m, time: 1.250000 sec"),
        ],
    ),
    SingleCaseParams(
        name="error-AC-short",
        inbytes=b"Pi\n",
        outbytes=b"3.1415926536\n",
        error=1e-11,
        mock_measure=OjExecInfo(
            answer=b"3.14159265358\n", elapsed=1.25, memory=None, returncode=0
        ),
        expected=make_result(
            name="error-AC-short",
            elapsed=1.25,
            exitcode=0,
            memory=None,
            status=JudgeStatus.AC,
        ),
        expected_log=[
            log_output("error-AC-short: start"),
            log_output("error-AC-short: \x1b[32mAC\x1b[39m, time: 1.250000 sec"),
        ],
    ),
    SingleCaseParams(
        name="error-small-AC-abs",
        inbytes=b"Planck mass\n",
        outbytes=b"0.0000000217640\n",
        error=1e-12,
        mock_measure=OjExecInfo(
            answer=b"0.0000000217647\n", elapsed=1.25, memory=None, returncode=0
        ),
        expected=make_result(
            name="error-small-AC-abs",
            elapsed=1.25,
            exitcode=0,
            memory=None,
            status=JudgeStatus.AC,
        ),
        expected_log=[
            log_output("error-small-AC-abs: start"),
            log_output("error-small-AC-abs: \x1b[32mAC\x1b[39m, time: 1.250000 sec"),
        ],
    ),
    SingleCaseParams(
        name="error-small-AC-abs-short",
        inbytes=b"Planck mass\n",
        outbytes=b"0.000000021764\n",
        error=1e-12,
        mock_measure=OjExecInfo(
            answer=b"0.0000000217647\n", elapsed=1.25, memory=None, returncode=0
        ),
        expected=make_result(
            name="error-small-AC-abs-short",
            elapsed=1.25,
            exitcode=0,
            memory=None,
            status=JudgeStatus.AC,
        ),
        expected_log=[
            log_output("error-small-AC-abs-short: start"),
            log_output(
                "error-small-AC-abs-short: \x1b[32mAC\x1b[39m, time: 1.250000 sec"
            ),
        ],
    ),
    SingleCaseParams(
        name="error-small-WA-abs",
        inbytes=b"Planck mass\n",
        outbytes=b"0.0000000217640\n",
        error=1e-13,
        mock_measure=OjExecInfo(
            answer=b"0.0000000217647\n", elapsed=1.25, memory=None, returncode=0
        ),
        expected=make_result(
            name="error-small-WA-abs",
            elapsed=1.25,
            exitcode=0,
            memory=None,
            status=JudgeStatus.WA,
        ),
        expected_log=[
            log_output("error-small-WA-abs: start"),
            log_output(
                "error-small-WA-abs:input:\n"
                "\x1b[1mPlanck\x1b[0m\x1b[2m_\x1b[0m\x1b[1mmass\x1b[0m\x1b[2m\\n\x1b[0m",
            ),
            log_output(
                "error-small-WA-abs:answer:\n\x1b[1m0.0000000217647\x1b[0m\x1b[2m\\n\x1b[0m",
            ),
            log_output(
                "error-small-WA-abs:expected:\n\x1b[1m0.0000000217640\x1b[0m\x1b[2m\\n\x1b[0m",
            ),
            log_output("error-small-WA-abs: \x1b[31mWA\x1b[39m, time: 1.250000 sec"),
        ],
    ),
    SingleCaseParams(
        name="error-large-AC-rel",
        inbytes=b"Planck temp\n",
        outbytes=b"141678400000000000000000000000000\n",
        error=1e-12,
        mock_measure=OjExecInfo(
            answer=b"141678400000000000000000000000000\n",
            elapsed=1.25,
            memory=None,
            returncode=0,
        ),
        expected=make_result(
            name="error-large-AC-rel",
            elapsed=1.25,
            exitcode=0,
            memory=None,
            status=JudgeStatus.AC,
        ),
        expected_log=[
            log_output("error-large-AC-rel: start"),
            log_output("error-large-AC-rel: \x1b[32mAC\x1b[39m, time: 1.250000 sec"),
        ],
    ),
    SingleCaseParams(
        name="error-large-AC-rel-diff",
        inbytes=b"Planck temp\n",
        outbytes=b"141678400000000000000000000000000\n",
        error=1e-12,
        mock_measure=OjExecInfo(
            answer=b"141678400000021987654321987654321\n",
            elapsed=1.25,
            memory=None,
            returncode=0,
        ),
        expected=make_result(
            name="error-large-AC-rel-diff",
            elapsed=1.25,
            exitcode=0,
            memory=None,
            status=JudgeStatus.AC,
        ),
        expected_log=[
            log_output("error-large-AC-rel-diff: start"),
            log_output(
                "error-large-AC-rel-diff: \x1b[32mAC\x1b[39m, time: 1.250000 sec"
            ),
        ],
    ),
    SingleCaseParams(
        name="error-large-WA-rel",
        inbytes=b"Planck temp\n",
        outbytes=b"141678400000000000000000000000000\n",
        error=1e-13,
        mock_measure=OjExecInfo(
            answer=b"141678400000021987654321987654321\n",
            elapsed=1.25,
            memory=None,
            returncode=0,
        ),
        expected=make_result(
            name="error-large-WA-rel",
            elapsed=1.25,
            exitcode=0,
            memory=None,
            status=JudgeStatus.WA,
        ),
        expected_log=[
            log_output("error-large-WA-rel: start"),
            log_output(
                "error-large-WA-rel:input:\n"
                "\x1b[1mPlanck\x1b[0m\x1b[2m_\x1b[0m\x1b[1mtemp\x1b[0m\x1b[2m\\n\x1b[0m",
            ),
            log_output(
                "error-large-WA-rel:answer:\n\x1b[1m141678400000021987654321987654321\x1b[0m\x1b[2m\\n\x1b[0m",
            ),
            log_output(
                "error-large-WA-rel:expected:\n\x1b[1m141678400000000000000000000000000\x1b[0m\x1b[2m\\n\x1b[0m",
            ),
            log_output("error-large-WA-rel: \x1b[31mWA\x1b[39m, time: 1.250000 sec"),
        ],
    ),
    SingleCaseParams(
        name="lines-same",
        inbytes=b"abc\n",
        outbytes=b"ABC\nDEF\nGH\n",
        mock_measure=OjExecInfo(
            answer=b"ABC\nDEF\nGH\n", elapsed=1.25, memory=None, returncode=0
        ),
        expected=make_result(
            name="lines-same",
            elapsed=1.25,
            exitcode=0,
            memory=None,
            status=JudgeStatus.AC,
        ),
        expected_log=[
            log_output("lines-same: start"),
            log_output("lines-same: \x1b[32mAC\x1b[39m, time: 1.250000 sec"),
        ],
    ),
    SingleCaseParams(
        name="lines-crlf",
        inbytes=b"abc\n",
        outbytes=b"ABC\nDEF\nGH\n",
        mock_measure=OjExecInfo(
            answer=b"ABC\r\nDEF\r\nGH\r\n", elapsed=1.25, memory=None, returncode=0
        ),
        expected=make_result(
            name="lines-crlf",
            elapsed=1.25,
            exitcode=0,
            memory=None,
            status=JudgeStatus.AC,
        ),
        expected_log=[
            log_output("lines-crlf: start"),
            log_output("lines-crlf: \x1b[32mAC\x1b[39m, time: 1.250000 sec"),
        ],
    ),
    SingleCaseParams(
        name="lines-diff",
        inbytes=b"abc\n",
        outbytes=b"ABC\nDEF\nGH\n",
        mock_measure=OjExecInfo(
            answer=b"ABC\nDDF\nGH\n", elapsed=1.25, memory=None, returncode=0
        ),
        expected=make_result(
            name="lines-diff",
            elapsed=1.25,
            exitcode=0,
            memory=None,
            status=JudgeStatus.WA,
        ),
        expected_log=[
            log_output("lines-diff: start"),
            log_output("lines-diff:input:\n\x1b[1mabc\x1b[0m\x1b[2m\\n\x1b[0m"),
            log_output(
                "lines-diff:answer:\n"
                "\x1b[1mABC\x1b[0m\x1b[2m\\n\x1b[0m\x1b[1mDDF\x1b[0m\x1b[2m\\n\x1b[0m\x1b[1mGH\x1b[0m\x1b[2m\\n\x1b[0m",
            ),
            log_output(
                "lines-diff:expected:\n"
                "\x1b[1mABC\x1b[0m\x1b[2m\\n\x1b[0m\x1b[1mDEF\x1b[0m\x1b[2m\\n\x1b[0m\x1b[1mGH\x1b[0m\x1b[2m\\n\x1b[0m",
            ),
            log_output("lines-diff: \x1b[31mWA\x1b[39m, time: 1.250000 sec"),
        ],
    ),
    SingleCaseParams(
        name="lines-error-diff-len",
        inbytes=b"abc\n",
        outbytes=b"ABC\nDEF\n3.14159265358979\n",
        error=1e-5,
        mock_measure=OjExecInfo(
            answer=b"ABC\nDEF\n3.14159265\n", elapsed=1.25, memory=None, returncode=0
        ),
        expected=make_result(
            name="lines-error-diff-len",
            elapsed=1.25,
            exitcode=0,
            memory=None,
            status=JudgeStatus.AC,
        ),
        expected_log=[
            log_output("lines-error-diff-len: start"),
            log_output("lines-error-diff-len: \x1b[32mAC\x1b[39m, time: 1.250000 sec"),
        ],
    ),
    SingleCaseParams(
        name="lines-error-grid",
        inbytes=b"abc\n",
        outbytes=b"ABC CD\nDEF 3.141592653589\n3.14159265358979\n",
        error=1e-5,
        mock_measure=OjExecInfo(
            answer=b"ABC CD\nDEF 3.14159265\n3.14159265\n",
            elapsed=1.25,
            memory=None,
            returncode=0,
        ),
        expected=make_result(
            name="lines-error-grid",
            elapsed=1.25,
            exitcode=0,
            memory=None,
            status=JudgeStatus.AC,
        ),
        expected_log=[
            log_output("lines-error-grid: start"),
            log_output("lines-error-grid: \x1b[32mAC\x1b[39m, time: 1.250000 sec"),
        ],
    ),
    SingleCaseParams(
        name="lines-grid-diff",
        inbytes=b"abc\n",
        outbytes=b"ABC CD\nDEF 3.141592653589\n3.14159265358979\n",
        error=1e-5,
        mock_measure=OjExecInfo(
            answer=b"ABC CD\nDEF 3.14159265 3.14\n3.14159265\n",
            elapsed=1.25,
            memory=None,
            returncode=0,
        ),
        expected=make_result(
            name="lines-grid-diff",
            elapsed=1.25,
            exitcode=0,
            memory=None,
            status=JudgeStatus.WA,
        ),
        expected_log=[
            log_output("lines-grid-diff: start"),
            log_output("lines-grid-diff:input:\n\x1b[1mabc\x1b[0m\x1b[2m\\n\x1b[0m"),
            log_output(
                "lines-grid-diff:answer:\n"
                "\x1b[1mABC\x1b[0m\x1b[2m_\x1b[0m\x1b[1mCD\x1b[0m\x1b[2m\\n"
                "\x1b[0m\x1b[1mDEF\x1b[0m\x1b[2m_\x1b[0m\x1b[1m3.14159265\x1b[0m\x1b[2m_\x1b[0m\x1b[1m3.14\x1b[0m\x1b[2m\\n"
                "\x1b[0m\x1b[1m3.14159265\x1b[0m\x1b[2m\\n\x1b[0m",
            ),
            log_output(
                "lines-grid-diff:expected:\n"
                "\x1b[1mABC\x1b[0m\x1b[2m_\x1b[0m\x1b[1mCD\x1b[0m\x1b[2m\\n"
                "\x1b[0m\x1b[1mDEF\x1b[0m\x1b[2m_\x1b[0m\x1b[1m3.141592653589\x1b[0m\x1b[2m\\n"
                "\x1b[0m\x1b[1m3.14159265358979\x1b[0m\x1b[2m\\n\x1b[0m",
            ),
            log_output("lines-grid-diff: \x1b[31mWA\x1b[39m, time: 1.250000 sec"),
        ],
    ),
    SingleCaseParams(
        name="lines-grid-diff-lines",
        inbytes=b"abc\n",
        outbytes=b"ABC CD\nDEF 3.141592653589\n3.14159265358979\n",
        error=1e-5,
        mock_measure=OjExecInfo(
            answer=b"ABC CD\nDEF 3.14159265 3.14\n3.14159265\nno\n",
            elapsed=1.25,
            memory=None,
            returncode=0,
        ),
        expected=make_result(
            name="lines-grid-diff-lines",
            elapsed=1.25,
            exitcode=0,
            memory=None,
            status=JudgeStatus.WA,
        ),
        expected_log=[
            log_output("lines-grid-diff-lines: start"),
            log_output(
                "lines-grid-diff-lines:input:\n\x1b[1mabc\x1b[0m\x1b[2m\\n\x1b[0m"
            ),
            log_output(
                "lines-grid-diff-lines:answer:\n"
                "\x1b[1mABC\x1b[0m\x1b[2m_\x1b[0m\x1b[1mCD\x1b[0m\x1b[2m\\n"
                "\x1b[0m\x1b[1mDEF\x1b[0m\x1b[2m_\x1b[0m\x1b[1m3.14159265\x1b[0m\x1b[2m_\x1b[0m\x1b[1m3.14\x1b[0m\x1b[2m\\n"
                "\x1b[0m\x1b[1m3.14159265\x1b[0m\x1b[2m\\n\x1b[0m\x1b[1mno\x1b[0m\x1b[2m\\n\x1b[0m",
            ),
            log_output(
                "lines-grid-diff-lines:expected:\n"
                "\x1b[1mABC\x1b[0m\x1b[2m_\x1b[0m\x1b[1mCD\x1b[0m\x1b[2m\\n"
                "\x1b[0m\x1b[1mDEF\x1b[0m\x1b[2m_\x1b[0m\x1b[1m3.141592653589\x1b[0m\x1b[2m\\n"
                "\x1b[0m\x1b[1m3.14159265358979\x1b[0m\x1b[2m\\n\x1b[0m",
            ),
            log_output("lines-grid-diff-lines: \x1b[31mWA\x1b[39m, time: 1.250000 sec"),
        ],
    ),
    SingleCaseParams(
        name="error-str-same",
        inbytes=b"abc\n",
        outbytes=b"ABC\n",
        error=1e-3,
        mock_measure=OjExecInfo(
            answer=b"ABC\n", elapsed=1.25, memory=None, returncode=0
        ),
        expected=make_result(
            name="error-str-same",
            elapsed=1.25,
            exitcode=0,
            memory=None,
            status=JudgeStatus.AC,
        ),
        expected_log=[
            log_output("error-str-same: start"),
            log_output("error-str-same: \x1b[32mAC\x1b[39m, time: 1.250000 sec"),
        ],
    ),
    SingleCaseParams(
        name="error-str-WA",
        inbytes=b"abc\n",
        outbytes=b"DEF\n",
        error=1e-3,
        mock_measure=OjExecInfo(
            answer=b"ABC\n", elapsed=1.25, memory=None, returncode=0
        ),
        expected=make_result(
            name="error-str-WA",
            elapsed=1.25,
            exitcode=0,
            memory=None,
            status=JudgeStatus.WA,
        ),
        expected_log=[
            log_output("error-str-WA: start"),
            log_output("error-str-WA:input:\n\x1b[1mabc\x1b[0m\x1b[2m\\n\x1b[0m"),
            log_output("error-str-WA:answer:\n\x1b[1mABC\x1b[0m\x1b[2m\\n\x1b[0m"),
            log_output("error-str-WA:expected:\n\x1b[1mDEF\x1b[0m\x1b[2m\\n\x1b[0m"),
            log_output("error-str-WA: \x1b[31mWA\x1b[39m, time: 1.250000 sec"),
        ],
    ),
    SingleCaseParams(
        name="TLE",
        inbytes=b"abc\n",
        outbytes=b"ABC\n",
        mock_measure=OjExecInfo(
            answer=b"ABC\n", elapsed=1.25, memory=None, returncode=None
        ),
        expected=make_result(
            name="TLE",
            elapsed=1.25,
            exitcode=None,
            memory=None,
            status=JudgeStatus.TLE,
        ),
        expected_log=[
            log_output("TLE: start"),
            log_output("TLE:input:\n\x1b[1mabc\x1b[0m\x1b[2m\\n\x1b[0m"),
            log_output("TLE:expected:\n\x1b[1mABC\x1b[0m\x1b[2m\\n\x1b[0m"),
            log_output("TLE: \x1b[31mTLE\x1b[39m, time: 1.250000 sec"),
        ],
    ),
    SingleCaseParams(
        name="RE",
        inbytes=b"abc\n",
        outbytes=b"ABC\n",
        mock_measure=OjExecInfo(
            answer=b"ABC\n", elapsed=1.25, memory=None, returncode=1
        ),
        expected=make_result(
            name="RE",
            elapsed=1.25,
            exitcode=1,
            memory=None,
            status=JudgeStatus.RE,
        ),
        expected_log=[
            log_output("RE: start"),
            log_output("RE:input:\n\x1b[1mabc\x1b[0m\x1b[2m\\n\x1b[0m"),
            log_output("RE:expected:\n\x1b[1mABC\x1b[0m\x1b[2m\\n\x1b[0m"),
            log_output("RE: \x1b[31mRE\x1b[39m, time: 1.250000 sec, return code: 1"),
        ],
    ),
    SingleCaseParams(
        name="MLE-NotMeasure",
        inbytes=b"abc\n",
        outbytes=b"ABC\n",
        mle=128,
        mock_measure=OjExecInfo(
            answer=b"ABC\n", elapsed=1.25, memory=None, returncode=0
        ),
        expected=make_result(
            name="MLE-NotMeasure",
            elapsed=1.25,
            exitcode=0,
            memory=None,
            status=JudgeStatus.AC,
        ),
        expected_log=[
            log_output("MLE-NotMeasure: start"),
            log_output("MLE-NotMeasure: \x1b[32mAC\x1b[39m, time: 1.250000 sec"),
        ],
    ),
    SingleCaseParams(
        name="MLE-Safe",
        inbytes=b"abc\n",
        outbytes=b"ABC\n",
        mle=128,
        mock_measure=OjExecInfo(
            answer=b"ABC\n", elapsed=1.25, memory=128, returncode=0
        ),
        expected=make_result(
            name="MLE-Safe",
            elapsed=1.25,
            exitcode=0,
            memory=128,
            status=JudgeStatus.AC,
        ),
        expected_log=[
            log_output("MLE-Safe: start"),
            log_output(
                "MLE-Safe: \x1b[32mAC\x1b[39m, time: 1.250000 sec, memory: 128.000000 MB"
            ),
        ],
    ),
    SingleCaseParams(
        name="MLE",
        inbytes=b"abc\n",
        outbytes=b"ABC\n",
        mle=128,
        mock_measure=OjExecInfo(
            answer=b"ABC\n", elapsed=1.25, memory=128.1, returncode=1
        ),
        expected=make_result(
            name="MLE",
            elapsed=1.25,
            exitcode=1,
            memory=128.1,
            status=JudgeStatus.MLE,
        ),
        expected_log=[
            log_output("MLE: start"),
            log_output("MLE:input:\n\x1b[1mabc\x1b[0m\x1b[2m\\n\x1b[0m"),
            log_output("MLE:answer:\n\x1b[1mABC\x1b[0m\x1b[2m\\n\x1b[0m"),
            log_output("MLE:expected:\n\x1b[1mABC\x1b[0m\x1b[2m\\n\x1b[0m"),
            log_output(
                "MLE: \x1b[31mMLE\x1b[39m,"
                " time: 1.250000 sec, memory: 128.100000 MB, return code: 1"
            ),
        ],
    ),
    SingleCaseParams(
        name="unicode-error",
        inbytes=b"a\n",
        outbytes=b"A\n",
        mock_measure=OjExecInfo(
            answer=b"\x82\xa0\n", elapsed=1.25, memory=None, returncode=0
        ),
        expected=make_result(
            name="unicode-error",
            elapsed=1.25,
            exitcode=0,
            memory=None,
            status=JudgeStatus.WA,
        ),
        expected_log=[
            log_output("unicode-error: start"),
            log_output("unicode-error:input:\n\x1b[1ma\x1b[0m\x1b[2m\\n\x1b[0m"),
            log_output(
                "unicode-error:answer:\n"
                "\x1b[2m'utf-8' codec can't decode byte 0x82 in position 0: invalid "
                "start byte\x1b[0m\x1b[1m\ufffd\ufffd\x1b[0m\x1b[2m\\n\x1b[0m"
            ),
            log_output("unicode-error:expected:\n\x1b[1mA\x1b[0m\x1b[2m\\n\x1b[0m"),
            log_output("unicode-error: \x1b[31mWA\x1b[39m, time: 1.250000 sec"),
        ],
    ),
    SingleCaseParams(
        name="long-lines",
        inbytes=b"abc\n",
        outbytes=b"ABC\n",
        mock_measure=OjExecInfo(
            answer=b"""ea226088-3cc7-4d33-985c-00004dfcdccb
6f791930-4b08-4618-a110-32be65332f08
87f28765-03bb-4aa3-ad2c-5e781aa60ca0
2c006877-aeae-4b24-88f1-29351ac40ad7
73380cec-9775-4b5d-9f86-b0bba61f2608
""",
            elapsed=1.25,
            memory=None,
            returncode=0,
        ),
        expected=make_result(
            name="long-lines",
            elapsed=1.25,
            exitcode=0,
            memory=None,
            status=JudgeStatus.WA,
        ),
        expected_log=[
            log_output("long-lines: start"),
            log_output("long-lines:input:\n\x1b[1mabc\x1b[0m\x1b[2m\\n\x1b[0m"),
            log_output(
                "long-lines:answer:\n"
                "\x1b[1mea226088-3cc7-4d33-9\x1b[0m\x1b[2m... (145 chars) "
                "...\x1b[0m\x1b[1md-9f86-b0bba61f2608\x1b[0m\x1b[2m\\n\x1b[0m",
            ),
            log_output("long-lines:expected:\n\x1b[1mABC\x1b[0m\x1b[2m\\n\x1b[0m"),
            log_output("long-lines: \x1b[31mWA\x1b[39m, time: 1.250000 sec"),
        ],
    ),
    SingleCaseParams(
        name="long-text",
        inbytes=b"abc\n",
        outbytes=b"ABC\n",
        mock_measure=OjExecInfo(
            answer=b"ea226088-3cc7-4d33-985c-00004dfcdccb6f791930-4b08-4618-a110-32be65332f08"
            b"87f28765-03bb-4aa3-ad2c-5e781aa60ca09b063b0f-7b48-4a1d-a49f-4e8524ace9174"
            b"2e6d61b-9656-4cb6-be7a-1a5d80224390961bed54-a32d-47c7-bb82-ae4591ed8ccd\r\n",
            elapsed=1.25,
            memory=None,
            returncode=0,
        ),
        expected=make_result(
            name="long-text",
            elapsed=1.25,
            exitcode=0,
            memory=None,
            status=JudgeStatus.WA,
        ),
        expected_log=[
            log_output("long-text: start"),
            log_output(
                "long-text:input:\n\x1b[1mabc\x1b[0m\x1b[2m\\n\x1b[0m",
            ),
            log_output(
                "long-text:answer:\n"
                "\x1b[1mea226088-3cc7-4d33-9\x1b[0m\x1b[2m... (178 chars) ...\x1b[0m"
                "\x1b[1m-bb82-ae4591ed8ccd\x1b[0m\x1b[2m\\r\\n\x1b[0m",
            ),
            log_output("long-text:expected:\n\x1b[1mABC\x1b[0m\x1b[2m\\n\x1b[0m"),
            log_output("long-text: \x1b[31mWA\x1b[39m, time: 1.250000 sec"),
        ],
    ),
    SingleCaseParams(
        name="long-lines-break",
        inbytes=b"abc\n",
        outbytes=b"ABC\n",
        mock_measure=OjExecInfo(
            answer=b"""ea226088-3cc7-4d33-\n985c-00004dfcdccb
6f791930-4b08-4618-a110-32be65332f08
87f28765-03bb-4aa3-ad2c-5e781aa60ca0
2c006877-aeae-4b24-88f1-29351ac40ad7
73380cec-9775-4b5d\n-9f86-b0bba61f2608
""",
            elapsed=1.25,
            memory=None,
            returncode=0,
        ),
        expected=make_result(
            name="long-lines-break",
            elapsed=1.25,
            exitcode=0,
            memory=None,
            status=JudgeStatus.WA,
        ),
        expected_log=[
            log_output("long-lines-break: start"),
            log_output("long-lines-break:input:\n\x1b[1mabc\x1b[0m\x1b[2m\\n\x1b[0m"),
            log_output(
                "long-lines-break:answer:\n"
                "\x1b[1mea226088-3cc7-4d33-\x1b[0m\x1b[2m\\n\x1b[0m\x1b[2m... (147 chars) "
                "...\x1b[0m\x1b[2m\\n\x1b[0m\x1b[1m-9f86-b0bba61f2608\x1b[0m\x1b[2m\\n\x1b[0m",
            ),
            log_output(
                "long-lines-break:expected:\n\x1b[1mABC\x1b[0m\x1b[2m\\n\x1b[0m"
            ),
            log_output("long-lines-break: \x1b[31mWA\x1b[39m, time: 1.250000 sec"),
        ],
    ),
]


@pytest.mark.usefixtures("mock_generate_test_cases")
@pytest.mark.parametrize(
    SingleCaseParams._fields,
    test_single_case_params,
    ids=[f"{i}:{t.name}" for i, t in enumerate(test_single_case_params)],
    indirect=["mock_measure", "mock_judge"],
)
def test_single_case(
    name: str,
    inbytes: bytes,
    outbytes: bytes | None,
    error: float | None,
    mle: float | None,
    expected: OjTestcaseResult,
    expected_log: list[tuple[str, int, str]],
    caplog: pytest.LogCaptureFixture,
    mock_judge: Problem,
    mock_measure: OjExecInfo,
    testtemp: pathlib.Path,
):
    caplog.set_level(logging.NOTSET)
    input_path = testtemp / "infile"
    input_path.write_bytes(inbytes)

    if outbytes is None:
        output_path = None
    else:
        output_path = testtemp / "outfile"
        output_path.write_bytes(outbytes)

    result = single_case(
        name,
        test_input_path=input_path,
        test_output_path=output_path,
        args=OjTestArguments(
            command="ls ~",
            problem=mock_judge,
            error=error,
            mle=mle,
            tle=None,
        ),
    )

    expected.answer = mock_measure.answer or b""
    expected.input = input_path
    expected.expected = output_path
    assert result == expected
    assert caplog.record_tuples == expected_log


def test_compare_answer_too_large_error(caplog: pytest.LogCaptureFixture):
    assert compare_answer(b"1", b"2", error=1)
    assert caplog.record_tuples == []

    assert compare_answer(b"1", b"2", error=1.0001)
    assert caplog.record_tuples == [
        (
            "competitive_verifier.oj.oj_test",
            logging.WARNING,
            "the tolerance is too large: relative = 1.0001",
        ),
    ]


test_oj_test_params: dict[str, tuple[dict[str, Any], OjTestArguments]] = {
    "default": (
        {
            "problem": LibraryCheckerProblem(problem_id="example"),
            "command": "ls .",
            "tle": 2,
            "error": None,
            "mle": 128,
            "env": None,
        },
        OjTestArguments(
            command="ls .",
            tle=2,
            mle=128,
            error=None,
            problem=LibraryCheckerProblem(problem_id="example"),
        ),
    ),
    "with_env": (
        {
            "problem": YukicoderProblem(problem_no=10),
            "command": ["ls", "."],
            "tle": 30,
            "error": None,
            "mle": 256,
            "env": {"TOKEN": "Dummy"},
        },
        OjTestArguments(
            command=["ls", "."],
            tle=30,
            mle=256,
            error=None,
            problem=YukicoderProblem(problem_no=10),
            env={"TOKEN": "Dummy"},
        ),
    ),
}


@pytest.mark.parametrize(
    ("args", "expected"),
    test_oj_test_params.values(),
    ids=test_oj_test_params.keys(),
)
def test_oj_test(
    mocker: MockerFixture,
    args: dict[str, Any],
    expected: OjTestArguments,
):
    mocker.patch(
        "competitive_verifier.oj.problem.LibraryCheckerProblem.checker_exe_name",
        "mockcheck",
    )
    run = mocker.patch("competitive_verifier.oj.oj_test._run")

    oj.test(**args)

    run.assert_called_once_with(expected)



import pathlib
from collections.abc import Iterable, Iterator
from logging import getLogger
from typing import TypeVar

from competitive_verifier.models import TestCaseData, TestCaseFile

logger = getLogger(__name__)


def _name_to_filename(name: str, ext: str):
    return pathlib.Path(name).with_suffix(f".{ext}").name


def save_testcases(samples: Iterable[TestCaseData], *, directory: pathlib.Path):
    for sample in samples:
        save_testcase(sample, directory=directory)


def save_testcase(sample: TestCaseData, *, directory: pathlib.Path):
    for data, ext in [(sample.input_data, "in"), (sample.output_data, "out")]:
        path = directory / _name_to_filename(sample.name, ext)

        if path.exists():
            logger.error("Failed to download since file already exists: %s", path)
        path.parent.mkdir(parents=True, exist_ok=True)
        path.write_bytes(data)
        logger.debug("saved to: %s", path)


_InOut = TypeVar("_InOut")


def enumerate_inouts(
    inputs: dict[str, _InOut],
    outputs: dict[str, _InOut],
) -> Iterator[tuple[str, _InOut, _InOut]]:
    common_keys = inputs.keys() & outputs.keys()
    if len(inputs) != len(common_keys) or len(outputs) != len(common_keys):
        logger.warning("dangling output case")

    if len(common_keys) == 0:
        logger.warning("no cases found")

    for key in sorted(common_keys):
        yield (key, inputs[key], outputs[key])


def merge_testcase_files(
    inputs: dict[str, pathlib.Path],
    outputs: dict[str, pathlib.Path],
) -> Iterator[TestCaseFile]:
    for name, i, o in enumerate_inouts(inputs, outputs):
        yield TestCaseFile(name=name, input_path=i, output_path=o)


def _casename(path: pathlib.Path, *, directory: pathlib.Path) -> str:
    return path.relative_to(directory).with_suffix("").as_posix()


def iter_testcases(
    *, directory: pathlib.Path, recursive: bool = False
) -> Iterator[TestCaseFile]:
    inputs: dict[str, pathlib.Path] = {}
    outputs: dict[str, pathlib.Path] = {}
    pre = "**/" if recursive else ""

    for path in directory.glob(pre + "*.in"):
        if path.is_file():
            inputs[_casename(path, directory=directory)] = path
    for path in directory.glob(pre + "*.out"):
        if path.is_file():
            outputs[_casename(path, directory=directory)] = path

    return merge_testcase_files(inputs, outputs)
